
# simple_grammar.py

from simple_lexer import Lexer

# --------------------------------------------------------------------------

class Rule (object) :
   def __init__ (self) :
       self.name = ""
       self.expr = None

class Expression (object) :
   def __init__ (self) :
       self.alternatives = [ ]

class Alternative (object) :
   def __init__ (self) :
       self.items = [ ]

class Ebnf (object) :
   def __init__ (self) :
       self.mark = ""
       self.expr = None
       self.impl = None

class Nonterminal (object) :
   def __init__ (self) :
       self.variable = ""
       self.rule = ""

class Terminal (object) :
   def __init__ (self) :
       self.text = ""

# --------------------------------------------------------------------------

class Grammar (Lexer) :

   def __init__ (self) :
       super (Grammar, self).__init__ ()
       self.rules = [ ]

   def skipDirective (self) :
       while self.isSeparator ('<') :
          self.nextToken ()
          while not self.isEndOfSource () and not self.isSeparator ('>') :
             self.nextToken ()
          if self.isEndOfSource () :
             self.error ("Unfinished directive")
          self.nextToken () # skip '>'

   def readIdentifier (self, msg) :
       name = ""
       if not self.isIdentifier () :
          self.error (msg)
       else :
          name = self.tokenText
          self.nextToken ()
       return name

   def parseRule (self) :
       rule = Rule ()

       if self.token != self.identifier :
          self.error ("Rule idenitifier expected")
       rule.name = self.tokenText
       rule.line = self.tokenLineNum
       self.nextToken ()

       self.skipDirective ()

       self.checkSeparator (':')
       rule.expr = self.parseExpression ()
       self.checkSeparator (';')

       self.rules.append (rule)

   def parseExpression (self) :
       expr = Expression ()
       expr.alternatives.append (self.parseAlternative ())
       while self.isSeparator ('|') :
          self.nextToken ()
          expr.alternatives.append (self.parseAlternative ())
       return expr

   def parseAlternative (self) :
       alt = Alternative ()
       self.skipDirective ()
       while not self.isSeparator ('|') and not self.isSeparator (')') and not self.isSeparator (';') :
          if self.token == self.identifier :
             item = self.parseNonterminal ()
             alt.items.append (item)
          elif self.token == self.character or self.token == self.string :
             item = self.parseTerminal ()
             alt.items.append (item)
          elif self.isSeparator ('(') :
             item = self.parseEbnf ()
             alt.items.append (item)
          else:
             self.error ("Unknown grammar item")
          self.skipDirective ()
       return alt

   def parseEbnf (self) :
       item = Ebnf ()
       self.checkSeparator ('(')
       item.expr = self.parseExpression ()
       self.checkSeparator (')')

       if self.isSeparator ('?') :
          item.mark = '?'
          self.nextToken ()
       elif self.isSeparator ('+') :
          item.mark = '+'
          self.nextToken ()
       elif self.isSeparator ('*') :
          item.mark = '*'
          self.nextToken ()
       elif self.isSeparator ('=') :
          self.nextToken ();
          self.checkSeparator ('>')
          item.impl = self.parseAlternative ()

       return item

   def parseNonterminal (self) :
       item = Nonterminal ()

       if self.token != self.identifier :
          self.error ("identifier expected")
       name = self.tokenText
       self.nextToken ()

       if not self.isSeparator (':') :
          item.rule = name
       else :
          self.nextToken () # skip ':'
          if not self.token == self.identifier :
             self.error ("rule identifier expected")
          item.variable = name
          item.rule = self.tokenText
          self.nextToken ()

       return item

   def parseTerminal (self) :
       item = Terminal ()
       if self.token != self.character and self.token != self.string :
          self.error ("string expected")
       item.text = self.tokenText
       self.nextToken ()
       return item

   def parseRules (self) :
       while not self.isEndOfSource () :
          self.parseRule ()

# --------------------------------------------------------------------------

if __name__ == "__main__" :
    grammar = Grammar ()
    grammar.openFile ("pas.g")
    grammar.parseRules ()
    for rule in grammar.rules :
       print (rule.name)

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
